"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Any, Dict, List

import numpy as np
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.drawfunctions import draw_bboxes, draw_tags


class Node(AbstractNode):
    """This node draws large bounding boxes over multiple object bounding boxes
    which have been identified as belonging to large groups."""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)
        self.tag = config["tag"]
        self.bbox_thickness = config["bbox_thickness"]
        self.bbox_color = tuple(config["bbox_color"])

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ Draws large bounding boxes over multiple object bounding boxes
        which have been identified as belonging to large groups

        Args:
            inputs (dict): Dict with keys "img", "bboxes", "obj_groups",
                "large_groups".

        Returns:
            outputs (dict): Dict with keys "img".
        """

        group_bboxes = self._get_group_bbox_coords(
            inputs["large_groups"], inputs["bboxes"], inputs["obj_groups"])
        group_tags = self._get_group_tags(
            inputs["large_groups"], self.tag)

        draw_bboxes(inputs["img"], group_bboxes,
                    self.bbox_color, self.bbox_thickness)   # type: ignore
        draw_tags(inputs["img"], group_bboxes,
                  group_tags, self.bbox_color)  # type: ignore

        return {}

    @staticmethod
    def _get_group_bbox_coords(large_groups: List[int],
                               bboxes: List[np.array],
                               obj_groups: List[int]) -> List[np.array]:
        """ For bboxes that belong to the same large group, get the coordinates of
        a large bbox that combines all these individual bboxes. Repeat for all large
        groups.
        """
        group_bboxes = []
        bboxes = np.array(bboxes)
        obj_groups = np.array(obj_groups)
        for group in large_groups:
            # filter relevant bboxes, select top-left and bot-right corners
            group_bbox = np.array([1., 1., 0., 0.])
            selected_bboxes = bboxes[obj_groups == group]
            group_bbox[:2] = np.amin(selected_bboxes, axis=0)[:2]
            group_bbox[2:] = np.amax(selected_bboxes, axis=0)[2:]
            group_bboxes.append(group_bbox)

        return group_bboxes

    @staticmethod
    def _get_group_tags(large_groups: List[int], tag: str) -> List[str]:
        """ Creates a list of tags to be used for the draw_tags function.
        """

        group_tags = [tag for _ in range(len(large_groups))]

        return group_tags



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Dict, Any

import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Node that checks if any objects are near to each other"""
    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)
        self.near_thres = config["near_threshold"]
        self.tag_msg = config["tag_msg"]

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Compares the 3D locations of all objects to see which objects are close to each other.
        If an object is close to another, tag it.

        Args:
            inputs (dict): Dict with keys "obj_3D_locs".

        Returns:
            outputs (dict): Dict with keys "obj_tags".
        """

        obj_tags = [""]*len(inputs["obj_3D_locs"])

        for idx_1, loc_1 in enumerate(inputs["obj_3D_locs"]):
            for idx_2, loc_2 in enumerate(inputs["obj_3D_locs"]):
                if idx_1 == idx_2:
                    continue

                dist_bet = np.linalg.norm(loc_1 - loc_2)
                if dist_bet < self.near_thres:
                    obj_tags[idx_1] = self.tag_msg
                    break

        return {"obj_tags": obj_tags}



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any, Dict, List, Tuple

import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.heuristic.utils.quick_find import QuickFind


class Node(AbstractNode):
    """This node groups objects that are near to each other."""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)

        self.obj_dist_thres = config["obj_dist_thres"]

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ Checks the distance between 3d locations of a pair of objects.
        If distance is less than threshold, assign the objects to the same group.
        Repeat for all object pairs.

        Args:
            inputs (dict): Dict with keys "obj_3D_locs".

        Returns:
            outputs (dict): Dict with keys "obj_groups".
        """

        nearby_obj_pairs = self._find_nearby_obj_pairs(
            inputs["obj_3D_locs"], self.obj_dist_thres)

        quickfind = QuickFind(len(inputs["obj_3D_locs"]))
        for (idx_1, idx_2) in nearby_obj_pairs:
            if not quickfind.connected(idx_1, idx_2):
                quickfind.union(idx_1, idx_2)

        return {"obj_groups": quickfind.get_group_alloc()}

    @staticmethod
    def _find_nearby_obj_pairs(obj_locs: List[np.array],
                               obj_dist_thres: float) -> List[Tuple[int, int]]:
        """If the distance between 2 objects are less than the threshold, append their
        indexes to nearby_obj_pairs as a tuple."""

        nearby_obj_pairs = []
        for idx_1, loc_1 in enumerate(obj_locs):
            for idx_2, loc_2 in enumerate(obj_locs):
                if idx_1 == idx_2:
                    continue

                dist_bet = np.linalg.norm(loc_1 - loc_2)

                if dist_bet <= obj_dist_thres:
                    if (idx_2, idx_1) not in nearby_obj_pairs:
                        nearby_obj_pairs.append((idx_1, idx_2))

        return nearby_obj_pairs



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from typing import Any, Tuple
import numpy as np
import cv2

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def set_res(stream: Any, desired_width: int, desired_height: int) -> None:
    '''
    Sets the resolution for the video frame
    '''
    stream.set(cv2.CAP_PROP_FRAME_WIDTH, desired_width)
    stream.set(cv2.CAP_PROP_FRAME_HEIGHT, desired_height)
    actual_width, actual_height = get_res(stream)
    if desired_width != actual_width:
        logger.warning("Unable to change width of video frame to %s, current width: %s!",
                       desired_width, actual_width)
    if desired_height != actual_height:
        logger.warning("Unable to change height of video frame to %s, current height: %s!",
                       desired_height, actual_height)


def get_res(stream: Any) -> Tuple[int, int]:
    '''
    Gets the resolution for the video frame
    '''
    width = int(stream.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(stream.get(cv2.CAP_PROP_FRAME_HEIGHT))

    return width, height


def mirror(frame: np.array) -> np.array:
    '''
    Mirrors a video frame.
    '''
    return cv2.flip(frame, 1)



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import logging
from typing import List
from peekingduck.pipeline.pipeline import Pipeline
from peekingduck.loaders import DeclarativeLoader
from peekingduck.pipeline.nodes.node import AbstractNode

END_TYPE = 'process_end'

"""
Combine runner at this level. Use this to create the graph, and waddle is loop or once
"""


class Runner():
    """Runner class that uses the declared nodes to create pipeline to run inference
    """

    def __init__(self, RUN_PATH: str, CUSTOM_NODE_PATH: str = None,
                 nodes: List[AbstractNode] = None):
        """
        Args:
            RUN_PATH (str): path to yaml file of node pipeine declaration.
            CUSTOM_NODE_PATH (str): path to custom nodes folder if used.
            nodes (:obj:'list' of :obj:'Node'): if not using declarations via yaml,
                initialize by giving the node stack as a list

        """

        self.logger = logging.getLogger(__name__)

        if not nodes:

            # create Graph to run
            self.node_loader = DeclarativeLoader(
                RUN_PATH, CUSTOM_NODE_PATH)  # type: ignore

            self.pipeline = self.node_loader.get_nodes()

        # If Runner given nodes, instantiated_nodes is created differently
        else:
            try:
                self.pipeline = Pipeline(nodes)
            except ValueError as error:
                self.logger.error(str(error))
                sys.exit(1)

    def run(self) -> None:
        """execute single or continuous inference
        """
        while not self.pipeline.video_end:
            self.pipeline.execute()
        del self.pipeline

    def get_run_config(self) -> List[str]:
        """retrieve run configs

        Returns:
            Dict[Any]: run configs being used for runner
        """
        return self.node_loader.node_list



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, Any, Tuple
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """This node converts bounding boxes to a singel pt which is
    the bottom midpoint of the bounding box"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)
        self.img_size = None

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Converts bounding boxes to a single point of reference
        for use in zone analytics

        Args:
            inputs (dict): Dict with keys "bboxes" and "img".
        Returns:
            outputs (dict): Dict with keys "btm_midpoint".

        """
        # get xy midpoint of each bbox (x1, y1, x2, y2)
        # This is calculated as x is (x1-x2)/2 and y is y2
        bboxes = inputs["bboxes"]
        frame = inputs["img"]
        self.img_size = (frame.shape[1], frame.shape[0])  # type:ignore
        return {'btm_midpoint': [self._xy_on_img(((bbox[0]+bbox[2])/2), bbox[3])
                                 for bbox in bboxes]}

    def _xy_on_img(self, pt_x: float, pt_y: float) -> Tuple[int, int]:
        """Return the int x y points of the midpoint on the original image"""
        assert self.img_size is not None
        return (int(pt_x * self.img_size[0]), int(pt_y * self.img_size[1]))



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, Any

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Counting node class that counts total number of detected objects"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Counts bboxes of object chosen in the frame. Note that this method
        requires that the bbox returns all the same objects (for example, all people)

        Args:
            inputs (dict): Dict with keys "bboxes".

        Returns:
            outputs (dict): Dict with keys "count".
        """
        return {'count': len(inputs["bboxes"])}



"""Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from typing import Any, Dict
from peekingduck.pipeline.nodes.node import AbstractNode
from .utils.drawfunctions import draw_count


class Node(AbstractNode):
    """Node that draws object counting"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Method that draws the count on the top left corner of image

         Args:
             inputs (dict): Dict with keys "count" and "img".
         Returns:
             outputs (dict): None
         """
        draw_count(inputs["img"],
                   inputs["count"])

        return {}



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any, Dict, List
from collections import Counter

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """This node checks which groups have exceeded the group size threshold."""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)
        self.group_size_thres = config["group_size_thres"]

    def run(self, inputs: Dict[str, List[int]]) -> Dict[str, List[int]]:
        """ Checks which groups have exceeded the group size threshold,
        and returns a list of such groups.

        Args:
            inputs (dict): Dict with keys "obj_groups".

        Returns:
            outputs (dict): Dict with keys "large_groups".
        """

        group_counter = Counter(inputs["obj_groups"])
        large_groups = [group for group in group_counter if
                        group_counter[group] > self.group_size_thres]

        return {"large_groups": large_groups}



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import os
import importlib
import logging
from typing import Any, Dict, List

import yaml

from peekingduck.pipeline.pipeline import Pipeline
from peekingduck.pipeline.nodes.node import AbstractNode


class ConfigLoader:  # pylint: disable=too-few-public-methods
    """ Reads configuration and returns configuration required to
    create the nodes for the project
    """

    def __init__(self, basedir: str) -> None:
        self._basedir = basedir

    def _get_config_path(self, node: str) -> str:
        """ Based on the node, return the corresponding node config path """

        configs_folder = os.path.join(self._basedir, 'configs')
        node_type, node_name = node.split(".")
        node_name = node_name + ".yml"
        filepath = os.path.join(configs_folder, node_type, node_name)

        return filepath

    def get(self, node_name: str) -> Dict[str, Any]:
        """Get item from configuration read from the filepath,
        item refers to the node item configuration you are trying to get"""

        filepath = self._get_config_path(node_name)

        with open(filepath) as file:
            node_config = yaml.load(file, Loader=yaml.FullLoader)

        # some models require the knowledge of where the root is for loading
        node_config['root'] = self._basedir
        return node_config


class DeclarativeLoader:  # pylint: disable=too-few-public-methods
    """Uses the declarative run_config.yml to load pipelines or compiled configs"""

    def __init__(self,
                 run_config: str,
                 custom_folder_path: str = 'src/custom_nodes') -> None:
        self.logger = logging.getLogger(__name__)

        pkdbasedir = os.path.join(os.path.dirname(os.path.abspath(__file__)))

        self.config_loader = ConfigLoader(pkdbasedir)
        self.custom_config_loader = ConfigLoader(custom_folder_path)
        self.node_list = self._load_node_list(run_config)
        self.custom_folder_path = custom_folder_path

    def _load_node_list(self, run_config: str) -> List[str]:
        """Loads a list of nodes from run_config.yml"""
        with open(run_config) as node_yml:
            nodes = yaml.load(node_yml, Loader=yaml.FullLoader)['nodes']

        self.logger.info('Successfully loaded run_config file.')
        return nodes

    def _instantiate_nodes(self) -> List[AbstractNode]:
        """ Given a list of imported nodes, instantiate nodes"""
        instantiated_nodes = []

        for node_item in self.node_list:
            admend_default_config = None
            node_str = node_item

            if isinstance(node_item, dict):
                node_str = list(node_item.keys())[0]  # type: ignore
                admend_default_config = {key: d[key] for d in node_item[node_str] for key in d}

            node_str_split = node_str.split('.')

            msg = "Initialising " + node_str + " node..."
            self.logger.info(msg)

            if len(node_str_split) == 3:
                path_to_node = ".".join(self.custom_folder_path.split('/')[-1:]) + "."
                node_name = ".".join(node_str_split[-2:])

                instantiated_node = self._init_node(path_to_node,
                                                    node_name,
                                                    self.custom_config_loader,
                                                    admend_default_config)  # type: ignore
            else:
                path_to_node = 'peekingduck.pipeline.nodes.'

                instantiated_node = self._init_node(path_to_node,
                                                    node_str,
                                                    self.config_loader,
                                                    admend_default_config)  # type: ignore

            instantiated_nodes.append(instantiated_node)

        return instantiated_nodes

    def _init_node(self, path_to_node: str, node_name: str,
                   config_loader: ConfigLoader,
                   admend_default_config: Dict[str, Any]) -> AbstractNode:
        """ Import node to filepath and initialise node with config """

        node = importlib.import_module(path_to_node + node_name)
        config = config_loader.get(node_name)

        if admend_default_config is not None:
            config = self._edit_node_config(config, admend_default_config)

        return node.Node(config)  # type: ignore

    def _edit_node_config(self, config: Dict[str, Any],
                          admend_default_config: Dict[str, Any]) -> Dict[str, Any]:
        " Edit default node configuration"

        params = set(admend_default_config.keys()) - set(config.keys())
        for param in params:
            admend_default_config.pop(param)
            msg = "'" + param + "' is not a valid configurable parameter"
            self.logger.warning(msg)
        config.update(admend_default_config)

        return config

    def get_nodes(self) -> Pipeline:
        """Returns a compiled Pipeline for PeekingDuck runner to execute"""
        instantiated_nodes = self._instantiate_nodes()

        try:
            return Pipeline(instantiated_nodes)
        except ValueError as error:
            self.logger.error(str(error))
            sys.exit(1)



"""
 Copyright 2021 AI Singapore

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

      https://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

from typing import Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode
from .utils.drawfunctions import draw_pts


class Node(AbstractNode):
    """This draw node draws the bottom midpoints"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Methods that draws btm midpoint of bounding bboxes

         Args:
             inputs (dict): Dict with keys "bboxes".

         Returns:
             outputs (dict): Empty dictionary.
         """
        draw_pts(inputs["img"],
                 inputs["btm_midpoint"])

        return {}
